#include "gdbstub.h"

#include "cpu/rv32cpu.h"
#include "memory/memory.h"

#include <errno.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

/* Sockets */
#include <unistd.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#define INVALID_SOCKET -1
#define SOCKET_ERROR -1
typedef int socket_t;
typedef struct sockaddr_in sockaddr_in_t;
typedef struct sockaddr sockaddr_t;

#define GDBSTUB_PORT 1234
#define PACKET_BUFFER_SIZE 2048

socket_t gdbstub_server_socket;
socket_t gdb_socket;
pthread_t gdbstub_thread;
pthread_t gdbstub_watcher_thread;

void gdbstub_thread_gdb();
void gdbstub_cpu_watcher_thread();
static void gdbstub_handle_ctrlc();

/*
 * Receive a packet from client gdb
 * Ignores the ACK on the way, and only returns the packet (not initial '$' symbol)
 */
static void gdbstub_recv_packet(char* packet, ssize_t* size)
{
	char data;
	recv(gdb_socket, &data, 1, 0);

	// Start of packet symbol is '$'
	// Skip data that is not packet
	if(data != '$')
	{
		switch(data)
		{
			case '+':
				// '+' : ACK
				// We ignore ACK, so read next
				return gdbstub_recv_packet(packet, size);
				break;
			case 0x3:
				// 0x3 : COMMAND CTRL-C (interrupt execution)
				gdbstub_handle_ctrlc();
				return gdbstub_recv_packet(packet, size);
				break;
			default:
				// Ignore unknown data
				return gdbstub_recv_packet(packet, size);
				break;
		}
	}

	// Receive packet until end of packet '#' symbol
	size_t i = 0;
	do
	{
		recv(gdb_socket, packet + i, 1, 0);
		i++;
	} while(packet[i - 1] != '#');

	// Receive checksum after end of packet symbol
	recv(gdb_socket, packet + i, 1, 0);
	recv(gdb_socket, packet + i + 1, 1, 0);

	if(size) *size = i + 2;
}

/*
 * Send a packet to client gdb
 * Pass only the packet data as parameter to this function
 */
static void gdbstub_send_packet(char* packet, ssize_t size)
{
	// Compute packet checksum
	char end[4] = {0};
	uint8_t sum = 0;
	for(size_t i = 0; i < size; i++)
	{
		sum += packet[i];
	}
	snprintf(end, 4, "#%02x", sum);

	// Send ACK to last packet, and begin of packet symbol
	char begin[2] = {'+', '$'};
	send(gdb_socket, begin, 2, 0);

	// Send the actual data of packet
	send(gdb_socket, packet, size, 0);

	// Send 'end of packet' and checksum
	send(gdb_socket, end, 3, 0);
}

/*
 * Send a packet to client gdb saying we don't support asked feature
 */
static void gdbstub_send_unsupported()
{
	char unsupported[] = "+$#00";
	send(gdb_socket, unsupported, 5, 0);
}

/*
 * Starts a server that will listen for GDB connections
 */
void gdbstub_start()
{
	socket_t sock = socket(AF_INET, SOCK_STREAM, 0);

	// Set REUSEADDRESS/PORT so that we dont get "port already in use" on close
	setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, &(int){1}, sizeof(int));
	setsockopt(sock, SOL_SOCKET, SO_REUSEPORT, &(int){1}, sizeof(int));

	if(sock == INVALID_SOCKET)
	{
		fprintf(stderr, "Could not create gdbstub server socket: %s\n", strerror(errno));
		exit(EXIT_FAILURE);
	}
	sockaddr_in_t sin = {0};
	sin.sin_addr.s_addr = htonl(INADDR_ANY);
	sin.sin_family = AF_INET;
	sin.sin_port = htons(GDBSTUB_PORT);

	if(bind(sock, (sockaddr_t*) &sin, sizeof(sin)) == SOCKET_ERROR)
	{
		fprintf(stderr, "Could not bind gdbstub server socket: %s\n", strerror(errno));
		exit(EXIT_FAILURE);
	}

	if(listen(sock, 1) == SOCKET_ERROR)
	{
		fprintf(stderr, "Could not listen gdbstub server socket: %s\n", strerror(errno));
		exit(EXIT_FAILURE);
	}

	gdbstub_server_socket = sock;
}

/*
 * Gracefully close gdb socket
 */
void gdbstub_stop()
{
	close(gdb_socket);
	close(gdbstub_server_socket);
}

/*
 * Wait for a client (gdb) to connect to this server
 */
void gdbstub_wait_for_connection()
{
	sockaddr_in_t csin = {0};
	socket_t c_socket;
	socklen_t sinsize = sizeof(csin);

	c_socket = accept(gdbstub_server_socket, (sockaddr_t*) &csin, &sinsize);
	if(c_socket == INVALID_SOCKET)
	{
		fprintf(stderr, "Could not accept client on gdbstub server: %s\n", strerror(errno));
		exit(EXIT_FAILURE);
	}

	gdb_socket = c_socket;

	pthread_create(&gdbstub_thread, 0, (void*) gdbstub_thread_gdb, 0);
	pthread_create(&gdbstub_watcher_thread, 0, (void*) gdbstub_cpu_watcher_thread, 0);
}

void gdbstub_thread_gdb()
{
	while(1)
	{
		char packet[PACKET_BUFFER_SIZE];
		ssize_t packet_size;
		gdbstub_recv_packet(packet, &packet_size);
		// printf("<-: %.*s\n", (int) packet_size, packet);

		// Analyse packet to respond correctly
		if(packet[0] == '?')
		{
			// Halt Reason : Signal 05 (SIGTRAP)
			char* resp = "S05";
			gdbstub_send_packet(resp, 3);
		}
		else if(packet[0] == 'g')
		{
			// g : read all registers -> send back all registers
			char resp[32 * 8 + 8 + 1] = {0};

			// Obtain CPU0 mutex
			pthread_mutex_lock(&cpu0_mutex);

			// All general purpose registers in host byte order as chars
			for(size_t i = 0; i < 32; i++)
			{
				uint32_t value = htonl(cpu0->regs.x[i]);
				snprintf(resp + i * 8, 9, "%08x", value);
			}

			// PC register
			uint32_t pc = htonl(cpu0->pc);
			snprintf(resp + 32 * 8, 9, "%08x", pc);

			// Let go of CPU0 mutex
			pthread_mutex_unlock(&cpu0_mutex);

			// Final packet size, send packet
			size_t size = 32 * 8 + 8;
			gdbstub_send_packet(resp, size);
		}
		else if(packet[0] == 'G')
		{
			// G : write all registers -> read and set all registers

			// Obtain CPU0 mutex
			pthread_mutex_lock(&cpu0_mutex);

			// All general purpose registers in host byte order as chars
			for(size_t i = 1; i < 32; i++)
			{
				uint32_t value = 0;
				sscanf(packet + 1 + i * 8, "%08x", &value);
				value = ntohl(value);

				cpu0->regs.x[i] = value;
			}

			// PC register
			uint32_t pc = 0;
			sscanf(packet + 1 + 32 * 8, "%08x", &pc);
			pc = ntohl(pc);
			cpu0->pc = pc;

			// Let go of CPU0 Mutex
			pthread_mutex_unlock(&cpu0_mutex);

			gdbstub_send_packet("OK", 2);
		}
		else if(packet[0] == 'm')
		{
			// m : read memory at address,length
			uint32_t address;
			uint32_t length;
			sscanf(packet + 1, "%x,%x", &address, &length);

			char data[length * 2 + 1];
			for(size_t i = 0; i < length; i++)
			{
				uint32_t value = mem_read8(address + i);
				snprintf(data + i * 2, 3, "%02x", value);
			}

			gdbstub_send_packet(data, length * 2);
		}
		else if(packet[0] == 'M')
		{
			// M : write memory at address,length:data
			uint32_t address;
			uint32_t length;
			sscanf(packet + 1, "%x,%x:", &address, &length);

			uint32_t data_start = 1;
			while(*(packet + data_start) != ':')
				data_start++;
			data_start++;

			for(size_t i = 0; i < length; i++)
			{
				uint32_t value;
				sscanf(packet + data_start + i * 2, "%02x", &value);
				mem_write8(address + i, value);
			}

			gdbstub_send_packet("OK", 2);
		}
		else if(packet[0] == 's')
		{
			// s : single-step

			// Send an ACK
			send(gdb_socket, "+", 1, 0);

			// Continue simulation, for 1 tick
			pthread_mutex_lock(&cpu0_mutex);
			cpu0->sim_ticks_left = 1;
			pthread_mutex_unlock(&cpu0_mutex);
			pthread_cond_signal(&cpu0->sim_condition);
		}
		else if(packet[0] == 'c')
		{
			// c : continue until something happens

			// Send an ACK
			send(gdb_socket, "+", 1, 0);

			// Continue simulation
			pthread_mutex_lock(&cpu0_mutex);
			cpu0->sim_ticks_left = -1;
			pthread_mutex_unlock(&cpu0_mutex);
			pthread_cond_signal(&cpu0->sim_condition);
		}
		else gdbstub_send_unsupported();
	}
}

void gdbstub_cpu_watcher_thread()
{
	while(1)
	{
		pthread_mutex_lock(&cpu0_mutex);
		pthread_cond_wait(&cpu0->sim_condition, &cpu0_mutex);
		if(!cpu0->sim_ticks_left && cpu0->sim_ticks_done > 0)
		{
			// Send back halt reason
			// Halt Reason : Signal 05 (SIGTRAP)
			char* resp = "S05";
			gdbstub_send_packet(resp, 3);
		}
		pthread_mutex_unlock(&cpu0_mutex);
	}
}

/*
 * Handles the GDB CTRL+C (^C) command, that
 * should interrupt the simulation
 */
static void gdbstub_handle_ctrlc()
{
	// Halt the simulation
	pthread_mutex_lock(&cpu0_mutex);
	cpu0->sim_ticks_left = 0;
	pthread_mutex_unlock(&cpu0_mutex);
}
